#include "FTPLUG.h"
#include "GUI.h"



AudioEffect* createEffectInstance(audioMasterCallback audioMaster)
{
	return new FTPLUG(audioMaster);
}



FTPLUG::FTPLUG(audioMasterCallback audioMaster) : AudioEffectX(audioMaster, NUM_PROGRAMS, NUM_PARAMS)
{
	int row,pgm,chn;

	setNumInputs (NUM_INPUTS);
	setNumOutputs(NUM_OUTPUTS);
	setUniqueID  (PLUGIN_UID);

	isSynth();
	canProcessReplacing();
	programsAreChunks(true);

	MidiChannel=0;

	for(chn=0;chn<MIDI_CHANNELS_ALL;++chn)
	{
		MidiProgram       [chn]=0;
		MidiProgramPrev   [chn]=0;
		MidiPitchBend     [chn]=0;
		MidiPitchBendRange[chn]=2.0f;
	}

	memset(Program,0,sizeof(Program));

	for(pgm=0;pgm<NUM_PROGRAMS;++pgm)
	{
		sprintf(Program[pgm].Name,"%3.3u default",pgm+1);

		strncpy(Program[pgm].StrVol    ,"15",ENVELOPE_TEXT_LEN);
		strncpy(Program[pgm].StrArp    ,"0",ENVELOPE_TEXT_LEN);
		strncpy(Program[pgm].StrPitch  ,"0",ENVELOPE_TEXT_LEN);
		strncpy(Program[pgm].StrHiPitch,"0",ENVELOPE_TEXT_LEN);
		strncpy(Program[pgm].StrDuty   ,"0",ENVELOPE_TEXT_LEN);
	}

	pgm=0;
	row=0;

	while(1)
	{
		if(PresetData[row][0]<0x20) break;

		sprintf(Program[pgm].Name,"%3.3i %s",pgm+1,PresetData[row++]);

		Program[pgm].ChType   =atoi(PresetData[row++]);
		Program[pgm].MultiMode=atoi(PresetData[row++]);

		strncpy(Program[pgm].StrVol    ,PresetData[row++],ENVELOPE_TEXT_LEN);
		strncpy(Program[pgm].StrArp    ,PresetData[row++],ENVELOPE_TEXT_LEN);
		strncpy(Program[pgm].StrPitch  ,PresetData[row++],ENVELOPE_TEXT_LEN);
		strncpy(Program[pgm].StrHiPitch,PresetData[row++],ENVELOPE_TEXT_LEN);
		strncpy(Program[pgm].StrDuty   ,PresetData[row++],ENVELOPE_TEXT_LEN);

		++pgm;
	}

	CompileInstruments();

	for(chn=0;chn<SYNTH_CHANNELS;++chn)
	{
		SynthChannel[chn].note=-1;
		SynthChannel[chn].period_prev=-1;	
	}

	OutputGain=100;

	SysType=TVSYS_NTSC;
	Polyphony=SYNTH_CHANNELS;
	SampleRatePrev=-1;	//force sample rate update

	MidiQueue.clear();

	MidiRPNLSB=0;
	MidiRPNMSB=0;
	MidiDataLSB=0;
	MidiDataMSB=0;

	memset(MidiKeyState,0,sizeof(MidiKeyState));

	suspend();

	setEditor((AEffEditor*)new GUI(this));

	SetUpdateGUI();
}



FTPLUG::~FTPLUG()
{
}



void FTPLUG::setParameter(VstInt32 index,float value)
{
}



float FTPLUG::getParameter(VstInt32 index)
{
	return 0;
} 



void FTPLUG::getParameterName(VstInt32 index,char *label)
{
} 



void FTPLUG::getParameterDisplay(VstInt32 index,char *text)
{
} 



void FTPLUG::getParameterLabel(VstInt32 index,char *label)
{
} 



VstInt32 FTPLUG::getProgram(void)
{
	UpdateGUI();	//force GUI update if current program has been changed while processReplacing is not called (plugin is added but not connected in Psycle)

	return MidiProgram[MidiChannel];
}



void FTPLUG::setProgram(VstInt32 program)
{
	if(program>NUM_PROGRAMS-1) program=NUM_PROGRAMS-1;
	
	MidiProgram[MidiChannel]=program;
}



void FTPLUG::getProgramName(char* name)
{
	strcpy(name,Program[MidiProgram[MidiChannel]].Name); 
}



void FTPLUG::setProgramName(char* name)
{
	strncpy(Program[MidiProgram[MidiChannel]].Name,name,MAX_NAME_LEN);

	Program[MidiProgram[MidiChannel]].Name[MAX_NAME_LEN-1]='\0';
} 



VstInt32 FTPLUG::canDo(char* text)
{
	if(!strcmp(text,"receiveVstEvents"   )) return 1;
	if(!strcmp(text,"receiveVstMidiEvent")) return 1;

	return -1;
}



VstInt32 FTPLUG::getNumMidiInputChannels(void)
{
	return MIDI_CHANNELS_ALL;
}



VstInt32 FTPLUG::getNumMidiOutputChannels(void)
{
	return 0;//no MIDI output
}



VstInt32 FTPLUG::SaveStringChunk(char *str,float *dst)
{
	VstInt32 i;

	for(i=0;i<(VstInt32)strlen(str);++i) *dst++=str[i];

	*dst++=0;

	return (VstInt32)strlen(str)+1;
}



VstInt32 FTPLUG::LoadStringChunk(char *str,int max_length,float *src)
{
	VstInt32 c,ptr;

	ptr=0;

	while(1)
	{
		c=(unsigned int)*src++;
		
		str[ptr++]=c;

		if(!c) break;

		if(ptr==max_length-1)
		{
			str[ptr]=0;
			break;
		}
	}

	return ptr;
}



VstInt32 FTPLUG::SavePresetChunk(float *chunk)
{
	VstInt32 chn,ptr,pgm;

	ptr=0;

	chunk[ptr++]=DATA;

	chunk[ptr++]=MPRG; for(chn=0;chn<MIDI_CHANNELS_ALL;++chn) chunk[ptr++]=(float)MidiProgram[chn];

	for(pgm=0;pgm<NUM_PROGRAMS;++pgm)
	{
		chunk[ptr++]=PROG; chunk[ptr++]=(float)pgm;

		chunk[ptr++]=TYPE; chunk[ptr++]=(float)Program[pgm].ChType;
		chunk[ptr++]=MMOD; chunk[ptr++]=(float)Program[pgm].MultiMode;

		chunk[ptr++]=NAME; ptr+=SaveStringChunk(Program[pgm].Name,&chunk[ptr]);

		chunk[ptr++]=EVOL; ptr+=SaveStringChunk(Program[pgm].StrVol    ,&chunk[ptr]);
		chunk[ptr++]=EARP; ptr+=SaveStringChunk(Program[pgm].StrArp    ,&chunk[ptr]);
		chunk[ptr++]=EPTC; ptr+=SaveStringChunk(Program[pgm].StrPitch  ,&chunk[ptr]);
		chunk[ptr++]=EHPT; ptr+=SaveStringChunk(Program[pgm].StrHiPitch,&chunk[ptr]);
		chunk[ptr++]=EDTY; ptr+=SaveStringChunk(Program[pgm].StrDuty   ,&chunk[ptr]);
	}

	chunk[ptr++]=FSYS; chunk[ptr++]=(float)SysType;

	chunk[ptr++]=GAIN; chunk[ptr++]=(float)OutputGain;
	chunk[ptr++]=DONE;

	return ptr*sizeof(float);//size in bytes
}



void FTPLUG::LoadPresetChunk(float *chunk)
{
	VstInt32 chn,pgm;
	float tag;

	pgm=0;

	while(1)
	{
		tag=*chunk++;

		if(tag==DONE) break;

		if(tag==MPRG) for(chn=0;chn<MIDI_CHANNELS_ALL;++chn) MidiProgram[chn]=(int)*chunk++;

		if(tag==PROG) pgm=(int)*chunk++;

		if(tag==TYPE) Program[pgm].ChType   =(int)*chunk++;
		if(tag==MMOD) Program[pgm].MultiMode=(int)*chunk++;

		if(tag==NAME) chunk+=LoadStringChunk(Program[pgm].Name,MAX_NAME_LEN,chunk);

		if(tag==EVOL) chunk+=LoadStringChunk(Program[pgm].StrVol    ,ENVELOPE_TEXT_LEN,chunk);
		if(tag==EARP) chunk+=LoadStringChunk(Program[pgm].StrArp    ,ENVELOPE_TEXT_LEN,chunk);
		if(tag==EPTC) chunk+=LoadStringChunk(Program[pgm].StrPitch  ,ENVELOPE_TEXT_LEN,chunk);
		if(tag==EHPT) chunk+=LoadStringChunk(Program[pgm].StrHiPitch,ENVELOPE_TEXT_LEN,chunk);
		if(tag==EDTY) chunk+=LoadStringChunk(Program[pgm].StrDuty   ,ENVELOPE_TEXT_LEN,chunk);

		if(tag==FSYS) SysType=(VstInt32)*chunk++;

		if(tag==GAIN) OutputGain=(VstInt32)*chunk++;
	}

	CompileInstruments();
}



VstInt32 FTPLUG::getChunk(void** data,bool isPreset)
{
	int size;

	size=SavePresetChunk(Chunk);

	*data=Chunk;

	return size;
}



VstInt32 FTPLUG::setChunk(void* data,VstInt32 byteSize,bool isPreset)
{
	if(byteSize>sizeof(Chunk)) return 0;

	memcpy(Chunk,data,byteSize);

	LoadPresetChunk(Chunk);

	SetUpdateGUI();

	return 0;
}



void FTPLUG::SetUpdateGUI(void)
{
	UpdateGuiFlag=true;
}



void FTPLUG::UpdateGUI(void)
{
	VstInt32 chn;

	for(chn=0;chn<MIDI_CHANNELS_ALL;++chn)	//this is needed because setProgram repeatedly called by the host to obtain all program names
	{
		if(MidiProgramPrev[chn]!=MidiProgram[chn])
		{
			MidiProgramPrev[chn]=MidiProgram[chn];

			SetUpdateGUI();
		}
	}

	if(UpdateGuiFlag)
	{
		UpdateGuiFlag=false;

		if(editor) ((GUI*)editor)->SetUpdate(true);

		updateDisplay();
	}
}



const char* FTPLUG::CompileEnvelope(char *str,EnvelopeDataStruct *env)
{
	VstInt32 ptr,len,loop;
	VstInt32 c,num,prefix;
	char *err;

	env->len=0;
	env->loop_ptr=0;
	env->has_loop=false;
	env->error=true;	//error by default, cleared on successful parsing

	err=NULL;

	len=(VstInt32)strlen(str);
	ptr=0;

	loop=-1;			//loop not defined yet

	while(ptr<len)
	{
		while(str[ptr]==' ') ++ptr;	//skip spaces

		c=str[ptr];

		if(!c) break;

		if(c>='a'&&c<='z') c-=32;	//case insensitive

		prefix=PARSE_PREFIX_NONE;

		if(c=='-') prefix=PARSE_PREFIX_NEGATIVE;

		if(c=='|')	//loop marker
		{
			if(loop<0)
			{
				loop=env->len;
			}
			else
			{
				err="Error: Loop point redefinition";
				break;
			}

			++ptr;

			if((ptr<len&&str[ptr]!=' ')||(ptr>=2&&str[ptr-2]!=' '))
			{
				err="Error: Loop marker should be preceded and followed with spacing";
				break;
			}
		}
		else
		if(prefix>PARSE_PREFIX_NONE||(c>='0'&&c<='9'))	//parse number with an optional prefix
		{
			if(prefix>PARSE_PREFIX_NONE)
			{
				++ptr;

				c=str[ptr];

				if(!(c>='0'&&c<='9'))
				{
					err="Error Prefix is not followed by a number";
					break;
				}
			}

			num=0;

			while(ptr<len)
			{
				c=str[ptr];

				if(c>='a'&&c<='z') c-=32;

				if(c<=32||c=='|'||c=='L'||c=='*') break;

				++ptr;

				if(c>='0'&&c<='9')
				{
					num=num*10+(c-'0');
					continue;
				}

				err="Error: Unexpected character while parsing a number";
				break;
			}

			if(prefix==PARSE_PREFIX_NEGATIVE) num=-num;

			if(num<-ENVELOPE_NUM_RANGE) num=-ENVELOPE_NUM_RANGE;
			if(num> ENVELOPE_NUM_RANGE) num= ENVELOPE_NUM_RANGE;

			env->value[env->len++]=num;

			if(env->len>ENVELOPE_DATA_LEN)
			{
				err="Error: Compiled envelope data is too long";
				break;
			}
		}
		else
		{
			err="Error: Unexpected character while parsing string";
			break;
		}
	}

	if(!err&&env->len)	//compiled successfully
	{
		env->error=false;

		if(loop>=0)
		{
			env->has_loop=true;
			env->loop_ptr=loop;
		}

		if(env->loop_ptr>env->len-1) env->loop_ptr=env->len-1;
		if(env->loop_ptr<0) env->loop_ptr=0;
	}

	if(err||!env->len)	//make default envelope if error or empty line
	{
		env->len=1;
		env->loop_ptr=0;
		env->value[0]=0;
	}

	return err;
}



const char* FTPLUG::CompileInstruments(void)
{
	int pgm;
	const char* err;

	for(pgm=0;pgm<NUM_PROGRAMS;++pgm)
	{
		err=NULL;

		err=CompileEnvelope(Program[pgm].StrVol    ,&Program[pgm].EnvVol);     if(err) break;
		err=CompileEnvelope(Program[pgm].StrArp    ,&Program[pgm].EnvArp);     if(err) break;
		err=CompileEnvelope(Program[pgm].StrPitch  ,&Program[pgm].EnvPitch);   if(err) break;
		err=CompileEnvelope(Program[pgm].StrHiPitch,&Program[pgm].EnvHiPitch); if(err) break;
		err=CompileEnvelope(Program[pgm].StrDuty   ,&Program[pgm].EnvDuty);    if(err) break;
	}

	return err;
}



void FTPLUG::MidiAddNote(VstInt32 delta,VstInt32 channel,VstInt32 note,VstInt32 velocity)
{
	MidiQueueStruct entry;

	entry.type    =mEventTypeNote;
	entry.channel =channel;
	entry.delta   =delta;
	entry.note    =note;
	entry.velocity=velocity;//0 for key off

	MidiQueue.push_back(entry);
}



void FTPLUG::MidiAddPitchBend(VstInt32 delta,VstInt32 channel,float depth)
{
	MidiQueueStruct entry;

	entry.type   =mEventTypePitchBend;
	entry.channel=channel;
	entry.delta  =delta;
	entry.depth  =depth;

	MidiQueue.push_back(entry);
}



bool FTPLUG::MidiIsAnyKeyDown(void)
{
	int i;

	for(i=0;i<128;++i) if(MidiKeyState[i]) return true;

	return false;
}



void FTPLUG::MidiAddProgramChange(VstInt32 delta,VstInt32 channel,VstInt32 program)
{
	MidiQueueStruct entry;

	entry.type   =mEventTypeProgram;
	entry.channel=channel;
	entry.delta  =delta;
	entry.program=program;

	MidiQueue.push_back(entry);
}



void FTPLUG::MidiAddControl(VstInt32 delta,VstInt32 channel,VstInt32 value)
{
	MidiQueueStruct entry;

	entry.type   =mEventTypeControl;
	entry.channel=channel;
	entry.delta  =delta;
	entry.value  =value;

	MidiQueue.push_back(entry);
}



VstInt32 FTPLUG::processEvents(VstEvents* ev) 
{ 
	VstMidiEvent* event;
	VstInt32 i,channel,status,note,velocity,wheel;
	char* midiData;

	for(i=0;i<ev->numEvents;++i) 
	{ 
		event=(VstMidiEvent*)ev->events[i];

		if(event->type!=kVstMidiType) continue; 

		midiData=event->midiData;

		channel=midiData[0]&0x0f;
		status =midiData[0]&0xf0;

		switch(status)
		{
		case 0x80:	//note off
		case 0x90:	//note on
			{
				note=midiData[1]&0x7f; 

				velocity=midiData[2]&0x7f; 

				if(status==0x80) velocity=0; 

				MidiAddNote(event->deltaFrames,channel,note,velocity);
			} 
			break;

		case 0xb0:	//control change
			{ 
				if(midiData[1]==0x64) MidiRPNLSB =midiData[2]&0x7f;
				if(midiData[1]==0x65) MidiRPNMSB =midiData[2]&0x7f;
				if(midiData[1]==0x26) MidiDataLSB=midiData[2]&0x7f;

				if(midiData[1]==0x06)
				{
					MidiDataMSB=midiData[2]&0x7f;

					if(MidiRPNLSB==0&&MidiRPNMSB==0) MidiPitchBendRange[channel]=(float)MidiDataMSB*.5f;
				}

				if(midiData[1]>=0x7b)//all notes off and mono/poly mode changes that also requires to do all notes off
				{
					MidiAddNote(event->deltaFrames,channel,-1,0);
				}
			} 
			break;

		case 0xc0:	//program change
			{
				MidiAddProgramChange(event->deltaFrames,channel,midiData[1]&0x7f);
			}
			break;

		case 0xe0:	//pitch bend change
			{
				wheel=(midiData[1]&0x7f)|((midiData[2]&0x7f)<<7);

				MidiAddPitchBend(event->deltaFrames,channel,(float)(wheel-0x2000)*MidiPitchBendRange[channel]/8192.0f);
			}
			break;
		}
	} 

	return 1; 
} 



void FTPLUG::LoadFTISeq(File* f,char* str)
{
	char buf[16];
	int cnt,len,loop,relp,set;

	memset(str,0,ENVELOPE_TEXT_LEN);

	if(!f->GetByte())
	{
		strncpy(str,"0",ENVELOPE_TEXT_LEN);
		return;
	}

	len =f->GetDWord();
	loop=f->GetDWord();
	relp=f->GetDWord();
	set =f->GetDWord();
	cnt =0;

	while(cnt<len)
	{
		if(cnt==loop) strncat(str,"| ",ENVELOPE_TEXT_LEN);

		sprintf(buf,"%i ",(signed char)f->GetByte());

		strncat(str,buf,ENVELOPE_TEXT_LEN);

		++cnt;
	}
}



const char* FTPLUG::LoadFTI(const char* filename)
{
	File f;
	char buf[1024];
	int n,pgm,len,type;

	pgm=MidiProgram[MidiChannel];

	if(!f.Open(filename,false)) return "Can't open file";

	f.GetString(buf,6);

	if(strcmp(buf,"FTI2.4")!=0)
	{
		f.Close();

		return "Not a valid or supported FTI file (needs to be FTI2.4)";
	}

	type=f.GetByte();

	switch(type)
	{
	case 1: type=CHTYPE_2A03_PULSE; break;
	case 2: type=CHTYPE_VRC6_PULSE; break;
	default: type=-1;
	}

	if(type<0)
	{
		f.Close();

		return "Unsupported instrument type (only 2A03 and VRC6 supported)";
	}

	Program[pgm].ChType=type;

	len=f.GetDWord();

	f.GetString(buf,len);

	memset(Program[pgm].Name,0,MAX_NAME_LEN);
	sprintf(Program[pgm].Name,"%3.3u ",pgm+1);
	strncat(Program[pgm].Name,buf,len);
	Program[pgm].Name[MAX_NAME_LEN-1]=0;

	n=f.GetByte();

	if(n!=5) 
	{
		f.Close();

		return "Unexpected number of sequences (should be 5)";
	}

	LoadFTISeq(&f,Program[pgm].StrVol);
	LoadFTISeq(&f,Program[pgm].StrArp);
	LoadFTISeq(&f,Program[pgm].StrPitch);
	LoadFTISeq(&f,Program[pgm].StrHiPitch);
	LoadFTISeq(&f,Program[pgm].StrDuty);

	f.Close();

	return NULL;
}



void FTPLUG::SaveFTISeq(File* f,EnvelopeDataStruct* env)
{
	int i;

	if(env->len<2)	//inactive sequence
	{
		if(!env->value[0])
		{
			f->PutByte(0);

			return;
		}
	}

	f->PutByte(1);
	f->PutDWord(env->len);
	f->PutDWord(env->has_loop?env->loop_ptr:-1);
	f->PutDWord(-1);
	f->PutDWord(0);

	for(i=0;i<env->len;++i) f->PutByte(env->value[i]);
}



const char* FTPLUG::SaveFTI(const char* filename)
{
	File f;
	int pgm,type;

	pgm=MidiProgram[MidiChannel];

	if(CompileInstruments()) return "Instrument sequences should be all correct in order to save";
	
	if(!f.Open(filename,true)) return "Can't create file";

	f.PutString("FTI2.4");

	type=0;

	switch(Program[pgm].ChType)
	{
	case CHTYPE_2A03_PULSE:
	case CHTYPE_2A03_TRIANGLE:
	case CHTYPE_2A03_NOISE:
		type=1;
		break;

	case CHTYPE_VRC6_PULSE:
	case CHTYPE_VRC6_SAW:
		type=2;
		break;
	}

	f.PutByte(type);						//instrument type
	f.PutDWord((unsigned int)strlen(Program[pgm].Name));	//instrument name length
	f.PutString(Program[pgm].Name);			//instrument name string

	f.PutByte(5);							//number of sequences

	SaveFTISeq(&f,&Program[pgm].EnvVol);
	SaveFTISeq(&f,&Program[pgm].EnvArp);
	SaveFTISeq(&f,&Program[pgm].EnvPitch);
	SaveFTISeq(&f,&Program[pgm].EnvHiPitch);
	SaveFTISeq(&f,&Program[pgm].EnvDuty);

	//dummy DPCM section

	f.PutDWord(0);	//number of assigned keys
	f.PutDWord(0);	//number of samples

	f.Close();

	return NULL;
}



VstInt32 FTPLUG::SynthAllocateVoice(VstInt32 midi_ch,VstInt32 note)
{
	VstInt32 chn;

	if(Polyphony==1) return 0;	//always use ch0 in mono mode

	for(chn=0;chn<Polyphony;++chn) if(SynthChannel[chn].midi_ch==midi_ch&&SynthChannel[chn].note==note) return chn;
	for(chn=0;chn<Polyphony;++chn) if(SynthChannel[chn].note<0) return chn;

	return -1;
}



void FTPLUG::SynthChannelChangeNote(VstInt32 chn,VstInt32 midi_ch,VstInt32 note,VstInt32 velocity)
{
	VstInt32 pgm,key;
	ChannelStruct* ch;

	key=note;
	pgm=MidiProgram[midi_ch];

	if(Program[pgm].MultiMode)
	{
		pgm+=note%12;
		note=60;

		while(pgm<0) pgm+=NUM_PROGRAMS;
		while(pgm>=NUM_PROGRAMS) pgm-=NUM_PROGRAMS;
	}

	if(velocity>100) velocity=100;

	ch=&SynthChannel[chn];

	ch->midi_ch=midi_ch;
	ch->key=key;
	ch->note=note;
	ch->type=Program[pgm].ChType;

	if(velocity>=0)
	{
		ch->velocity=velocity;
		
		ch->ptr_vol=0;
		ch->ptr_arp=0;
		ch->ptr_pitch=0;
		ch->ptr_hipitch=0;
		ch->ptr_duty=0;

		ch->pitch_acc=0;
		ch->arp_offset=0;
		ch->vol_output=0;
		ch->duty_output=0;
		ch->period_prev=-1;

		ch->env_vol    =&Program[pgm].EnvVol;
		ch->env_arp    =&Program[pgm].EnvArp;
		ch->env_pitch  =&Program[pgm].EnvPitch;
		ch->env_hipitch=&Program[pgm].EnvHiPitch;
		ch->env_duty   =&Program[pgm].EnvDuty;
	}
}



void FTPLUG::SynthChannelReleaseNote(VstInt32 midi_ch,VstInt32 note)
{
	int chn;

	if(note>=0)
	{
		for(chn=0;chn<SYNTH_CHANNELS;++chn)
		{
			if(SynthChannel[chn].midi_ch==midi_ch&&SynthChannel[chn].key==note) SynthChannel[chn].note=-1;
		}
	}
	else
	{
		for(chn=0;chn<SYNTH_CHANNELS;++chn)
		{
			if(SynthChannel[chn].midi_ch==midi_ch||midi_ch<0) SynthChannel[chn].note=-1;
		}
	}
}



void FTPLUG::SynthChannelUpdate(VstInt32 chn)
{
	ChannelStruct* ch;
	VstInt32 period,volume,duty;
	float freq;

	ch=&SynthChannel[chn];

	if(ch->note<0) return;

	//process volume envelope

	ch->vol_output=ch->env_vol->value[ch->ptr_vol++];

	if(ch->ptr_vol>=ch->env_vol->len)
	{
		if(!ch->vol_output&&!ch->env_vol->has_loop) //terminate channel that is looped to zero volume at end
		{
			ch->note=-1;
		}
		else
		{
			if(ch->env_vol->has_loop) ch->ptr_vol=ch->env_vol->loop_ptr; else --ch->ptr_vol;
		}
	}

	//process arpeggio envelope

	ch->arp_offset=ch->env_arp->value[ch->ptr_arp++];

	if(ch->ptr_arp>=ch->env_arp->len)
	{
		if(ch->env_arp->has_loop) ch->ptr_arp=ch->env_arp->loop_ptr; else --ch->ptr_arp;
	}

	//process pitch envelope, does not repeat last value for unlooped sequences

	if(ch->ptr_pitch<ch->env_pitch->len) ch->pitch_acc+=ch->env_pitch->value[ch->ptr_pitch++];

	if(ch->ptr_pitch>=ch->env_pitch->len)
	{
		if(ch->env_pitch->has_loop) ch->ptr_pitch=ch->env_pitch->loop_ptr;
	}

	//process hi-pitch envelope, does not repeat last value for unlooped sequences

	if(ch->ptr_hipitch<ch->env_hipitch->len) ch->pitch_acc+=ch->env_hipitch->value[ch->ptr_hipitch++];

	if(ch->ptr_hipitch>=ch->env_hipitch->len)
	{
		if(ch->env_hipitch->has_loop) ch->ptr_hipitch=ch->env_hipitch->loop_ptr;
	}

	//process duty envelope

	ch->duty_output=ch->env_duty->value[ch->ptr_duty++];

	if(ch->ptr_duty>=ch->env_duty->len)
	{
		if(ch->env_duty->has_loop) ch->ptr_duty=ch->env_duty->loop_ptr; else --ch->ptr_duty;
	}
	
	//convert envelopes data into APU register writes

	volume=ch->vol_output*ch->velocity/100;

	if(volume<0) volume=0;
	if(volume>15) volume=15;

	switch(ch->type)
	{
	case CHTYPE_2A03_PULSE:
		{
			freq=(float)(ch->note-69)+MidiPitchBend[ch->midi_ch]+ch->arp_offset;	//gives frequency in hz

			period=(VstInt32)((float)SysClock/(16.0f*(440.0f*pow(2.0f,freq/12.0f))))-1;

			period+=ch->pitch_acc;

			if(period<0||period>0x7ff) volume=0;	//mute notes out of range
		}
		break;

	case CHTYPE_2A03_TRIANGLE:
		{
			freq=(float)(ch->note-69)+MidiPitchBend[ch->midi_ch]+ch->arp_offset;	//gives frequency in hz

			period=(VstInt32)((float)SysClock/(32.0f*(440.0f*pow(2.0f,freq/12.0f))))-1;

			period+=ch->pitch_acc;

			if(period<0||period>0x7ff) volume=0;	//mute notes out of range
		}
		break;

	case CHTYPE_2A03_NOISE:
		{
			period=~ch->note-4-ch->arp_offset-ch->pitch_acc;
		}
		break;

	case CHTYPE_VRC6_PULSE:
		{
			freq=(float)(ch->note-69)+MidiPitchBend[ch->midi_ch]+ch->arp_offset;	//gives frequency in hz

			period=(VstInt32)((float)SysClock/(16.0f*(440.0f*pow(2.0f,freq/12.0f))))-1;

			period+=ch->pitch_acc;

			if(period<0||period>0x7ff) volume=0;	//mute notes out of range
		}
		break;

	case CHTYPE_VRC6_SAW:
		{
			freq=(float)(ch->note-69)+MidiPitchBend[ch->midi_ch]+ch->arp_offset;	//gives frequency in hz

			period=(VstInt32)((float)SysClock/(14.0f*(440.0f*pow(2.0f,freq/12.0f))))-1;

			period+=ch->pitch_acc;

			if(period<0||period>0x7ff) volume=0;	//mute notes out of range
		}
		break;
	}

	if(ch->type==CHTYPE_VRC6_PULSE||ch->type==CHTYPE_VRC6_SAW)
	{
		APU[chn].write_register(0x4015,0x00);	//mute all APU channels
	}
	else
	{
		APU[chn].write_register(0x9000,0x00);	//mute all VRC6 channels
		APU[chn].write_register(0xa000,0x00);
		APU[chn].write_register(0xb000,0x00);
	}

	duty=ch->duty_output;

	switch(ch->type)
	{
	case CHTYPE_2A03_PULSE:
		{
			APU[chn].write_register(0x4015,0x01);						//enable pulse 1
			APU[chn].write_register(0x4000,0x30|((duty&3)<<6)|(volume&0x0f));//set duty and volume
			APU[chn].write_register(0x4001,0x08);						//no sweep, extend period range to 11-bit
			APU[chn].write_register(0x4002,period&0xff);				//set pitch lsb

			if(ch->period_prev<0||(ch->period_prev&0x0700)!=(period&0x0700))
			{
				ch->period_prev=period&0x0700;

				APU[chn].write_register(0x4003,(period>>8)&0x07);		//set pitch msb
			}
		}
		break;

	case CHTYPE_2A03_TRIANGLE:
		{
			APU[chn].write_register(0x4015,0x04);						//enable triangle
			APU[chn].write_register(0x4008,volume?0xff:0x00);			//enable count
			APU[chn].write_register(0x400a,period&0xff);				//set pitch lsb
			APU[chn].write_register(0x400b,(period>>8)&7);				//set pitch msb
		}
		break;

	case CHTYPE_2A03_NOISE:
		{
			APU[chn].write_register(0x4015,0x08);						//enable noise
			APU[chn].write_register(0x400c,0xf0|(volume&0x0f));			//set volume
			APU[chn].write_register(0x400e,(period&0x0f)|((duty&1)<<7));	//set period and mode
			APU[chn].write_register(0x400f,0x00);						//length counter
		}
		break;

	case CHTYPE_VRC6_PULSE:
		{
			APU[chn].write_register(0xb000,0);							//disable saw
			APU[chn].write_register(0x9000,(volume&0x0f)|((duty&7)<<4));//set volume and duty
			APU[chn].write_register(0x9001,(period&0xff));				//set pitch lsb
			APU[chn].write_register(0x9002,0x80|((period>>8)&0x0f));	//set pitch msb and enable channel
		}
		break;

	case CHTYPE_VRC6_SAW:
		{
			APU[chn].write_register(0x9000,0);							//disable pulse
			APU[chn].write_register(0xb000,(volume&0x0f)|((duty&7^3)<<4));//set volume (duty used as top bits of volume)
			APU[chn].write_register(0xb001,(period&0xff));				//set pitch lsb
			APU[chn].write_register(0xb002,0x80|((period>>8)&0x0f));	//set pitch msb and enable channel
		}
		break;
	}

	APU[chn].end_frame();
}



void FTPLUG::processReplacing(float **inputs,float **outputs,VstInt32 sampleFrames)
{
	float *outL=outputs[0];
	float *outR=outputs[1];
	float level,sample_rate;
	VstInt32 i,chn,note;
	blip_sample_t buf[1];

	sample_rate=(float)updateSampleRate();

	if(SysType==TVSYS_NTSC)
	{
		SysClock    =NTSC_CLOCK;
		SysFrameRate=NTSC_FPS;
	}
	else
	{
		SysClock    =PAL_CLOCK;
		SysFrameRate=PAL_FPS;
	}

	if(SampleRatePrev!=sample_rate)	//sample rate changed (or plugin just started)
	{
		SampleRatePrev=sample_rate;

		for(chn=0;chn<SYNTH_CHANNELS;++chn)
		{
			APU[chn].apu.reset();
			APU[chn].vrc6.reset();

			APU[chn].set_timings((long)sample_rate,SysClock,SysFrameRate);
		}
	}

	while(--sampleFrames>=0)
	{
		for(i=0;i<(VstInt32)MidiQueue.size();++i)
		{
			if(MidiQueue[i].delta>=0)
			{
				--MidiQueue[i].delta;

				if(MidiQueue[i].delta<0)//new message
				{
					switch(MidiQueue[i].type)
					{
					case mEventTypeNote:
						{
							if(MidiQueue[i].velocity)//key on
							{
								MidiKeyState[MidiQueue[i].note]=1;

								chn=SynthAllocateVoice(MidiQueue[i].channel,MidiQueue[i].note);

								if(chn>=0)
								{
									SynthChannelChangeNote(chn,MidiQueue[i].channel,MidiQueue[i].note,MidiQueue[i].velocity);
								}
							}
							else//key off
							{
								if(MidiQueue[i].note>=0)
								{
									MidiKeyState[MidiQueue[i].note]=0;

									SynthChannelReleaseNote(MidiQueue[i].channel,MidiQueue[i].note);

									if(Polyphony>1)
									{
										for(chn=0;chn<SYNTH_CHANNELS;++chn) if(SynthChannel[chn].note==MidiQueue[i].note) SynthChannelReleaseNote(MidiQueue[i].channel,MidiQueue[i].note);
									}
									else
									{
										for(note=127;note>=0;--note)
										{
											if(MidiKeyState[note])
											{
												SynthChannelChangeNote(0,MidiQueue[i].channel,note,-1);

												break;
											}
										}
									}
								}
								else
								{
									memset(MidiKeyState,0,sizeof(MidiKeyState));	//stop all channels

									SynthChannelReleaseNote(-1,-1);
								}
							}
						}
						break;

					case mEventTypePitchBend:
						{
							MidiPitchBend[MidiQueue[i].channel]=MidiQueue[i].depth;
						}
						break;

					case mEventTypeProgram:
						{
							if(MidiProgram[MidiQueue[i].channel]!=MidiQueue[i].program)
							{
								MidiProgram[MidiQueue[i].channel]=MidiQueue[i].program;

								SetUpdateGUI();
							}
						}
						break;

					case mEventTypeControl:
						{			
						}
						break;
					}
				}
			}
		}

		//generate sound, update channels if needed

		level=0;

		for(chn=0;chn<SYNTH_CHANNELS;++chn)
		{
			if(APU[chn].samples_avail()==0)
			{
				SynthChannelUpdate(chn);
			}

			if(APU[chn].samples_avail()>0)
			{
				APU[chn].read_samples(buf,1);

				level+=(float)buf[0]/32768.0f;
			}
		}

		level=level*(float)OutputGain/100.0f;

		if(level<-1.0f) level=-1.0f;
		if(level>1.0f) level=1.0f;

		(*outL++)=level;
		(*outR++)=level;
	}

	MidiQueue.clear();

	UpdateGUI();
}
